#pragma once

#include "CoreMinimal.h"
#include "Enums/MujocoEnums.h"
#include "MujocoActuator.generated.h"

USTRUCT(BlueprintType)
struct FMujocoActuatorV2
{
	GENERATED_BODY()

	// Common attributes for all actuators	

	/** Type of the actuator. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator")
	EMujocoActuatorType Type = EMujocoActuatorType::General;

	/** Group of the actuator. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (Attribute = "group"))
	TOptional<int32> Group;

	/** If true, the control input to this actuator is automatically clamped to ctrlrange at runtime. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (Attribute = "ctrllimited"))
	TOptional<ELimitType> CtrlLimited;

	/** If true, the force output of this actuator is automatically clamped to forcerange at runtime. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (Attribute = "forcelimited"))
	TOptional<ELimitType> ForceLimited;

	/** If true, the internal state (activation) associated with this actuator is automatically clamped to actrange at runtime. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (Attribute = "actlimited"))
	TOptional<ELimitType> ActLimited;

	/** Range for clamping the control input. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (Attribute = "ctrlrange"))
	TOptional<FVector2D> CtrlRange;

	/** Range for clamping the force output. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (Attribute = "forcerange"))
	TOptional<FVector2D> ForceRange;

	/** Range for clamping the activation state. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (Attribute = "actrange"))
	TOptional<FVector2D> ActRange;

	/** Range of feasible lengths of the actuator’s transmission. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (Attribute = "lengthrange"))
	TOptional<FVector2D> LengthRange;

	/** This attribute scales the length (and consequently moment arms, velocity and force) of the actuator, for all transmission types. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (Attribute = "gear"))
	TOptional<FMujocoGear> Gear;

	/** Used only for the slider-crank transmission type. Specifies the length of the connecting rod. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (Attribute = "cranklength"))
	TOptional<double> CrankLength;

	/** This and the next four attributes determine the type of actuator transmission. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (Attribute = "joint"))
	FName Joint = NAME_None;

	/** Identical to joint, except that for ball and free joints, the 3d rotation axis given by gear is defined in the parent frame. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (Attribute = "jointinparent"))
	FName JointInParent = NAME_None;

	/** This transmission can apply force and torque at a site. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (Attribute = "site"))
	FName Site = NAME_None;

	/** When using a site transmission, measure the translation and rotation w.r.t the frame of the refsite. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (Attribute = "refsite"))
	FName RefSite = NAME_None;

	/** This transmission can apply linear forces at contact points in the direction of the contact normal. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (Attribute = "body"))
	FName Body = NAME_None;

	/** If specified, the actuator acts on the given tendon. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (Attribute = "tendon"))
	FName Tendon = NAME_None;

	/** If specified, the actuator acts on a slider-crank mechanism which is implicitly determined by the actuator. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (Attribute = "cranksite"))
	FName CrankSite = NAME_None;

	/** Used only for the slider-crank transmission type. The specified site is the pin joining the slider and the connecting rod. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (Attribute = "slidersite"))
	FName SliderSite = NAME_None;

	/** Dimension of the activation state. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (Attribute = "actdim"))
	int32 ActDim = -1;

	/** Activation dynamics type for the actuator. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (Attribute = "dyntype"))
	TOptional<EDynType> DynType;

	/** Gain type. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (Attribute = "gaintype"))
	TOptional<EGainType> GainType;

	/** Bias type. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (Attribute = "biastype"))
	TOptional<EBiasType> BiasType;

	/** Activation dynamics parameters. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (Attribute = "dynprm"))
	TArray<double> DynPrm;

	/** Gain parameters. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (Attribute = "gainprm"))
	TArray<double> GainPrm;

	/** Bias parameters. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (Attribute = "biasprm"))
	TArray<double> BiasPrm;

	/** If true, force computation will use the next value of the activation variable rather than the current one. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (Attribute = "actearly", EditConditionHides, EditCondition = "Type == EMujocoActuatorType::General"))
	TOptional<bool> ActEarly;

	// ===== Custom properties for specific actuator types =====

	/** Position feedback gain. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (EditConditionHides, EditCondition = "Type == EMujocoActuatorType::General || Type == EMujocoActuatorType::Position || Type == EMujocoActuatorType::IntVelocity", Attribute = "kp"))
	TOptional<double> Kp;

	/** Damping applied by the actuator. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (EditConditionHides, EditCondition = "Type == EMujocoActuatorType::General || Type == EMujocoActuatorType::Position || Type == EMujocoActuatorType::Velocity || Type == EMujocoActuatorType::IntVelocity || Type == EMujocoActuatorType::Damper", Attribute = "kv"))
	TOptional<double> Kv;

	/** Damping applied by the actuator, using damping ratio units. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (EditConditionHides, EditCondition = "Type == EMujocoActuatorType::General || Type == EMujocoActuatorType::Position || Type == EMujocoActuatorType::IntVelocity", Attribute = "dampratio"))
	TOptional<double> DampingRatio;

	/** Time-constant of optional first-order filter. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (EditConditionHides, EditCondition = "Type == EMujocoActuatorType::General || Type == EMujocoActuatorType::Position || Type == EMujocoActuatorType::Cylinder", Attribute = "timeconst"))
	TOptional<double> TimeConst;

	/** Automatically set the actuator’s ctrlrange to match the transmission target’s range. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (EditConditionHides, EditCondition = "Type == EMujocoActuatorType::General || Type == EMujocoActuatorType::Position || Type == EMujocoActuatorType::IntVelocity", Attribute = "inheritrange"))
	TOptional<double> InheritRange;

	/** Area of the cylinder. This is used internally as actuator gain. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (EditConditionHides, EditCondition = "Type == EMujocoActuatorType::General || Type == EMujocoActuatorType::Cylinder", Attribute = "area"))
	TOptional<double> Area;

	/** Instead of area the user can specify diameter. If both are specified, diameter has precedence. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (EditConditionHides, EditCondition = "Type == EMujocoActuatorType::General || Type == EMujocoActuatorType::Cylinder", Attribute = "diameter"))
	TOptional<double> Diameter;

	/** Bias parameters, copied internally into biasprm. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (EditConditionHides, EditCondition = "Type == EMujocoActuatorType::General || Type == EMujocoActuatorType::Cylinder", Attribute = "bias"))
	TOptional<FVector> Bias;

	// -- Muscle actuator custom properties --

	/** Time constants for activation and de-activation dynamics. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (EditConditionHides, EditCondition = "Type == EMujocoActuatorType::General || Type == EMujocoActuatorType::Muscle", Attribute = "timeconst"))
	TOptional<FVector2D> MuscleTimeConsts;

	/** Width of smooth transition between activation and deactivation time constants. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (EditConditionHides, EditCondition = "Type == EMujocoActuatorType::General || Type == EMujocoActuatorType::Muscle", Attribute = "tausmooth"))
	TOptional<double> TauSmooth;

	/** Operating length range of the muscle, in units of L0. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (EditConditionHides, EditCondition = "Type == EMujocoActuatorType::General || Type == EMujocoActuatorType::Muscle", Attribute = "range"))
	TOptional<FVector2D> Range;

	/** Peak active force at rest. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (EditConditionHides, EditCondition = "Type == EMujocoActuatorType::General || Type == EMujocoActuatorType::Muscle", Attribute = "force"))
	TOptional<double> Force;

	/** If the force attribute is negative, the peak active force for the muscle is set to this value divided by mjModel.actuator_acc0. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (EditConditionHides, EditCondition = "Type == EMujocoActuatorType::General || Type == EMujocoActuatorType::Muscle", Attribute = "scale"))
	TOptional<double> Scale;

	/** Shortening velocity at which muscle force drops to zero, in units of L0 per second. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (EditConditionHides, EditCondition = "Type == EMujocoActuatorType::General || Type == EMujocoActuatorType::Muscle", Attribute = "vmax"))
	TOptional<double> VMax;

	/** Passive force generated at lmax, relative to the peak rest force. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (EditConditionHides, EditCondition = "Type == EMujocoActuatorType::General || Type == EMujocoActuatorType::Muscle", Attribute = "fpmax"))
	TOptional<double> FpMax;

	/** Active force generated at saturating lengthening velocity, relative to the peak rest force. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (EditConditionHides, EditCondition = "Type == EMujocoActuatorType::General || Type == EMujocoActuatorType::Muscle", Attribute = "fvmax"))
	TOptional<double> FvMax;

	/** Lower position range of the normalized FLV curve, in units of L0. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (EditConditionHides, EditCondition = "Type == EMujocoActuatorType::General || Type == EMujocoActuatorType::Muscle", Attribute = "lmin"))
	TOptional<double> LMin;

	/** Upper position range of the normalized FLV curve, in units of L0. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (EditConditionHides, EditCondition = "Type == EMujocoActuatorType::General || Type == EMujocoActuatorType::Muscle", Attribute = "lmax"))
	TOptional<double> LMax;

	// -- Adhesion actuator custom properties --

	/** Gain of the adhesion actuator, in units of force. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco | Actuator", meta = (EditConditionHides, EditCondition = "Type == EMujocoActuatorType::General || Type == EMujocoActuatorType::Adhesion", Attribute = "gain"))
	TOptional<double> AdhesionGain;
};