#pragma once

#include "CoreMinimal.h"
#include "Enums/MujocoEnums.h"
#include "MujocoJoint.generated.h"

USTRUCT(BlueprintType)
struct FMujocoJoint
{
	GENERATED_BODY()

	/** Type of the joint. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco", meta = (Attribute = "type"))
	EMujocoJointType Type = EMujocoJointType::Hinge;	

	/** Integer group to which the joint belongs. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco", meta = (EditConditionHides, EditCondition = "Type != EMujocoJointType::None", Attribute = "group"))
	TOptional<int32> Group;

	/** Position of the joint, specified in the frame of the body where the joint is defined. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco", meta = (EditConditionHides, EditCondition = "Type != EMujocoJointType::None", Attribute = "pos"))
	TOptional<FVector> Position;

	/** Axis of rotation for hinge joints and direction of translation for slide joints. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco", meta = (EditConditionHides, EditCondition = "Type != EMujocoJointType::None", Attribute = "axis"))
	TOptional<FVector> Axis;

	/** Time constant and damping ratio for the mass-spring-damper system. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco", meta = (EditConditionHides, EditCondition = "Type != EMujocoJointType::None", Attribute = "springdamper"))
	TOptional<FVector2D> SpringDamper;

	/** Joint stiffness. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco", meta = (EditConditionHides, EditCondition = "Type != EMujocoJointType::None", Attribute = "stiffness"))
	TOptional<double> Stiffness;

	/** The joint limits. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco", meta = (EditConditionHides, EditCondition = "Type != EMujocoJointType::None", Attribute = "range"))
	TOptional<FVector2D> Range;

	/** Specifies if the joint has limits. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco", meta = (EditConditionHides, EditCondition = "Type != EMujocoJointType::None", Attribute = "limited"))
	TOptional<ELimitType> Limited;

	/** Range for clamping total actuator forces acting on this joint. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco", meta = (EditConditionHides, EditCondition = "Type != EMujocoJointType::None", Attribute = "actuatorfrcrange"))
	TOptional<FVector2D> ActuatorForceRange;

	/** Specifies whether actuator forces acting on the joint should be clamped. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco", meta = (EditConditionHides, EditCondition = "Type != EMujocoJointType::None", Attribute = "actuatorfrclimited"))
	TOptional<ELimitType> ActuatorForceLimitType;

	/** If true, gravity compensation applied to this joint is added to actuator forces. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco", meta = (EditConditionHides, EditCondition = "Type != EMujocoJointType::None", Attribute = "actuatorgravcomp"))
	TOptional<bool> bActuatorGravityCompensation;

	/** The distance threshold below which limits become active. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco", meta = (EditConditionHides, EditCondition = "Type != EMujocoJointType::None", Attribute = "margin"))
	TOptional<double> Margin;

	/** The reference position or angle of the joint. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco", meta = (EditConditionHides, EditCondition = "Type != EMujocoJointType::None", Attribute = "ref"))
	TOptional<double> Reference;

	/** The joint position or angle in which the joint spring achieves equilibrium. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco", meta = (EditConditionHides, EditCondition = "Type != EMujocoJointType::None", Attribute = "springref"))
	TOptional<double> SpringReference;

	/** Additional inertia associated with movement of the joint that is not due to body mass. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco", meta = (EditConditionHides, EditCondition = "Type != EMujocoJointType::None", Attribute = "armature"))
	TOptional<double> Armature;

	/** Damping applied to all degrees of freedom created by this joint. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco", meta = (EditConditionHides, EditCondition = "Type != EMujocoJointType::None", Attribute = "damping"))
	TOptional<double> Damping;

	/** Friction loss due to dry friction. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Mujoco", meta = (EditConditionHides, EditCondition = "Type != EMujocoJointType::None", Attribute = "frictionloss"))
	TOptional<double> FrictionLoss;
};
