#include "STLImportFactory.h"
#include "EngineUtils.h"
#include "Math/TransformVectorized.h"
#include "Engine/StaticMesh.h"
#include "MaterialDomain.h"
#include "StaticMeshDescription.h"

namespace UnrealSTL
{
	constexpr int32 BinaryHeaderSize = 80;
	constexpr int32 BinaryHeaderSizeAndSize = BinaryHeaderSize + 4;
	constexpr int32 BinaryTriangleSize = 50;
} // namespace UnrealSTL

FUnrealSTLMesh::FUnrealSTLMesh()
{
	TrianglesNum = 0;
}

FUnrealSTLMesh::FUnrealSTLMesh(const TArray<FUnrealSTLMesh>& Meshes) : FUnrealSTLMesh()
{
	FBox BoundingBox;
	BoundingBox.Init();

	for (const FUnrealSTLMesh& Mesh : Meshes)
	{
		Vertices += Mesh.Vertices;
		TrianglesNum += Mesh.TrianglesNum;
		BoundingBox += Mesh.Bounds.GetBox();
	}

	BoundingBox.GetCenterAndExtents(Bounds.Origin, Bounds.BoxExtent);
	Bounds.SphereRadius = 0;
	uint32 VertexBase = 0;
	for (const FUnrealSTLMesh& Mesh : Meshes)
	{
		FStaticMeshSection Section;
		Section.FirstIndex = LODIndices.Num();
		Section.NumTriangles = Mesh.TrianglesNum;
		Sections.Add(Section);
		for (const FStaticMeshBuildVertex& BuildVertex : Mesh.Vertices)
		{
			Bounds.SphereRadius = FMath::Max((BuildVertex.Position - FVector3f(Bounds.Origin)).Size(), Bounds.SphereRadius);
		}

		for (const uint32 VertexIndex : Mesh.LODIndices)
		{
			LODIndices.Add(VertexBase + VertexIndex);
		}
		VertexBase += Mesh.Vertices.Num();
	}
}

bool UUnrealSTLFactory::LoadMeshFromSTLData(FArrayReader& Reader, const FUnrealSTLConfig& Config, FUnrealSTLMesh& STLMesh)
{
	STLMesh.Vertices.Empty();
	STLMesh.LODIndices.Empty();
	STLMesh.TrianglesNum = 0;

	FBox BoundingBox;
	BoundingBox.Init();

	uint32 ProcessedTriangles = 0;

	Reader.Seek(0);

	if ((Config.FileMode == EUnrealSTLFileMode::Auto && !FMemory::Memcmp(Reader.GetData(), "solid", 5)) || Config.FileMode == EUnrealSTLFileMode::ASCII)
	{
		TArray<FString> Tokens;
		FString Token;
		for (int32 Offset = 0; Offset < Reader.Num(); Offset++)
		{
			char Char = static_cast<char>(Reader[Offset]);
			if (Char == 0 || Char == '\r' || Char == '\n' || Char == ' ' || Char == '\t')
			{
				if (!Token.IsEmpty())
				{
					Tokens.Add(Token);
					Token.Empty();
				}
			}
			else
			{
				Token += Char;
			}
		}

		int32 TokenIndex = 0;
		int32 VertexState = 0; // 0 to 2

		FStaticMeshBuildVertex Vertex[3];

		while (TokenIndex < Tokens.Num())
		{
			Token = Tokens[TokenIndex++];
			if (Token == "normal")
			{
				if (TokenIndex + 3 > Tokens.Num())
				{
					return false;
				}
				else
				{
					float NormalX = FCString::Atof(*Tokens[TokenIndex++]);
					float NormalY = FCString::Atof(*Tokens[TokenIndex++]);
					float NormalZ = FCString::Atof(*Tokens[TokenIndex++]);

					FVector TransformedNormal = Config.Transform.TransformVector(FVector(-NormalX, -NormalY, NormalZ));

					Vertex[0].TangentZ = FVector3f(TransformedNormal);
					Vertex[1].TangentZ = FVector3f(TransformedNormal);
					Vertex[2].TangentZ = FVector3f(TransformedNormal);
				}
			}
			else if (Token == "vertex")
			{
				if (TokenIndex + 3 > Tokens.Num())
				{
					return false;
				}
				else
				{
					float VertexX = FCString::Atof(*Tokens[TokenIndex++]);
					float VertexY = FCString::Atof(*Tokens[TokenIndex++]);
					float VertexZ = FCString::Atof(*Tokens[TokenIndex++]);

					FVector TransformedPosition = Config.Transform.TransformPosition(FVector(-VertexX, -VertexY, VertexZ));

					Vertex[VertexState++].Position = FVector3f(TransformedPosition);

					if (VertexState > 2)
					{
						VertexState = 0;
						BoundingBox += FVector(Vertex[0].Position);
						BoundingBox += FVector(Vertex[1].Position);
						BoundingBox += FVector(Vertex[2].Position);

						STLMesh.Vertices.Append(Vertex, 3);
						ProcessedTriangles++;
					}
				}
			}
		}
	}
	else
	{
		if (Reader.Num() < UnrealSTL::BinaryHeaderSizeAndSize)
		{
			return false;
		}

		Reader.Seek(UnrealSTL::BinaryHeaderSize);

		Reader << STLMesh.TrianglesNum;

		STLMesh.Vertices.Reserve(STLMesh.TrianglesNum * 3);

		while (ProcessedTriangles < STLMesh.TrianglesNum)
		{
			if (Reader.Tell() + UnrealSTL::BinaryTriangleSize <= Reader.Num())
			{
				FStaticMeshBuildVertex Vertex[3];

				FVector3f Normal;

				Reader << Normal;

				Vertex[0].TangentZ = Vertex[1].TangentZ = Vertex[2].TangentZ = Normal;

				Reader << Vertex[0].Position;
				Reader << Vertex[1].Position;
				Reader << Vertex[2].Position;

				for (int32 VertexIndex = 0; VertexIndex < 3; VertexIndex++)
				{
					Vertex[VertexIndex].Position = FVector3f(Config.Transform.TransformPosition(FVector(-Vertex[VertexIndex].Position.X, -Vertex[VertexIndex].Position.Y, Vertex[VertexIndex].Position.Z)));					
					Vertex[VertexIndex].TangentZ = FVector3f(Config.Transform.TransformVector(FVector(-Normal.X, -Normal.Y, Normal.Z)));
				}

				BoundingBox += FVector(Vertex[0].Position);
				BoundingBox += FVector(Vertex[1].Position);
				BoundingBox += FVector(Vertex[2].Position);

				STLMesh.Vertices.Append(Vertex, 3);

				uint16 AttributesBytesCount;
				Reader << AttributesBytesCount;

				if (AttributesBytesCount > 0)
				{
					int64 CurrentSeek = Reader.Tell();
					if (CurrentSeek + AttributesBytesCount <= Reader.Num())
					{
						Reader.Seek(CurrentSeek + AttributesBytesCount);
					}
					else
					{
						return false;
					}
				}
			}
			else
			{
				return false;
			}

			ProcessedTriangles++;
		}
	}

	BoundingBox.GetCenterAndExtents(STLMesh.Bounds.Origin, STLMesh.Bounds.BoxExtent);
	STLMesh.Bounds.SphereRadius = 0;
	int32 Index = 0;
	for (const FStaticMeshBuildVertex& BuildVertex : STLMesh.Vertices)
	{
		STLMesh.Bounds.SphereRadius = FMath::Max((BuildVertex.Position - FVector3f(STLMesh.Bounds.Origin)).Size(), STLMesh.Bounds.SphereRadius);
		if (Index % 3 == 0) // generate indices every 3 vertices
		{
			STLMesh.LODIndices.Add(Index);
			if (!Config.bReverseWinding)
			{
				STLMesh.LODIndices.Add(Index + 1);
				STLMesh.LODIndices.Add(Index + 2);
			}
			else
			{
				STLMesh.LODIndices.Add(Index + 2);
				STLMesh.LODIndices.Add(Index + 1);
			}
		}
		Index++;
	}

	STLMesh.TrianglesNum = ProcessedTriangles;

	return true;
}

UStaticMesh* UUnrealSTLFactory::LoadStaticMeshFromSTLFile(const FString& Filename, const FUnrealSTLConfig& Config, const FUnrealSTLStaticMeshConfig& StaticMeshConfig)
{
	FUnrealSTLFile STLFile;
	STLFile.Filename = Filename;
	STLFile.Config = Config;

	FUnrealSTLFileLOD STLFileLOD;
	STLFileLOD.Sections = { STLFile };
	STLFileLOD.ScreenSize = 1.0f;
	return LoadStaticMeshFromSTLFileLODs({ STLFileLOD }, StaticMeshConfig);
}

UStaticMesh* UUnrealSTLFactory::LoadStaticMeshFromSTLFileLODs(const TArray<FUnrealSTLFileLOD>& FileLODs, const FUnrealSTLStaticMeshConfig& StaticMeshConfig)
{
	if (FileLODs.Num() < 1)
	{
		return nullptr;
	}

	UStaticMesh* StaticMesh = NewObject<UStaticMesh>(StaticMeshConfig.Outer ? StaticMeshConfig.Outer : GetTransientPackage());

	StaticMesh->NeverStream = true;

#if WITH_EDITOR
	StaticMesh->bAutoComputeLODScreenSize = false;
#endif

	if (StaticMesh->GetRenderData())
	{
		StaticMesh->GetRenderData()->ReleaseResources();
	}

	StaticMesh->SetRenderData(MakeUnique<FStaticMeshRenderData>());
	FStaticMeshRenderData* RenderData = StaticMesh->GetRenderData();

	RenderData->AllocateLODResources(FileLODs.Num());

	int32 LODIndex = 0;
	TArray<FStaticMaterial> StaticMaterials;
	for (const FUnrealSTLFileLOD& FileLOD : FileLODs)
	{
		FStaticMeshLODResources& LODResources = RenderData->LODResources[LODIndex];
		RenderData->ScreenSize[LODIndex] = FileLOD.ScreenSize > 0 ? FileLOD.ScreenSize : (1.0f - ((1.f / FileLODs.Num()) * LODIndex));

#if WITH_EDITOR
		FStaticMeshSourceModel& SourceModel = StaticMesh->AddSourceModel();
		SourceModel.ScreenSize = RenderData->ScreenSize[LODIndex];
#endif

#if ENGINE_MAJOR_VERSION < 5
		FStaticMeshLODResources::FStaticMeshSectionArray& Sections = LODResources.Sections;
#else
		FStaticMeshSectionArray& Sections = LODResources.Sections;
#endif

		TArray<FUnrealSTLMesh> STLMeshes;
		for (const FUnrealSTLFile& File : FileLOD.Sections)
		{
			FArrayReader Data;
			if (!FFileHelper::LoadFileToArray(Data, *File.Filename))
			{
				return nullptr;
			}
			FUnrealSTLMesh STLMesh;
			if (!LoadMeshFromSTLData(Data, File.Config, STLMesh))
			{
				return nullptr;
			}
			STLMeshes.Add(MoveTemp(STLMesh));
		}

		// a bit of hacky optimizations...
		FUnrealSTLMesh STLMergedSections;
		FUnrealSTLMesh* STLSectionsPtr;
		if (STLMeshes.Num() == 1)
		{
			STLSectionsPtr = &STLMeshes[0];
		}
		else
		{
			STLMergedSections = FUnrealSTLMesh(STLMeshes);
			STLSectionsPtr = &STLMergedSections;
		}

		if (STLSectionsPtr->Sections.Num() == 0)
		{
			FStaticMeshSection DefaultSection;
			DefaultSection.NumTriangles = STLSectionsPtr->TrianglesNum;
			STLSectionsPtr->Sections.Add(DefaultSection);
		}

		for (int32 SectionIndex = 0; SectionIndex < STLSectionsPtr->Sections.Num(); SectionIndex++)
		{
			FStaticMeshSection& Section = Sections.AddDefaulted_GetRef();

			Section.FirstIndex = STLSectionsPtr->Sections[SectionIndex].FirstIndex;
			Section.NumTriangles = STLSectionsPtr->Sections[SectionIndex].NumTriangles;
			Section.MaterialIndex = StaticMaterials.Num();

#if WITH_EDITOR
			FMeshSectionInfo SectionInfo;
			SectionInfo.MaterialIndex = Section.MaterialIndex;
			StaticMesh->GetSectionInfoMap().Set(LODIndex, SectionIndex, SectionInfo);
#endif

			const FUnrealSTLConfig& Config = FileLOD.Sections[SectionIndex].Config;
			FStaticMaterial Material(Config.Material ? Config.Material : UMaterial::GetDefaultMaterial(MD_Surface), *FString::Printf(TEXT("LOD_%u_Section_%u"), LODIndex, SectionIndex));
			Material.UVChannelData.bInitialized = true;
			StaticMaterials.Add(Material);
		}

		LODResources.VertexBuffers.PositionVertexBuffer.Init(STLSectionsPtr->Vertices, StaticMeshConfig.bAllowCPUAccess);
		LODResources.VertexBuffers.StaticMeshVertexBuffer.Init(STLSectionsPtr->Vertices, 1, StaticMeshConfig.bAllowCPUAccess);
		if (StaticMeshConfig.bAllowCPUAccess)
		{
			LODResources.IndexBuffer = FRawStaticIndexBuffer(true);
		}
		LODResources.IndexBuffer.SetIndices(STLSectionsPtr->LODIndices, EIndexBufferStride::Force32Bit);

		if (LODIndex == 0)
		{
			RenderData->Bounds = STLSectionsPtr->Bounds;
		}

		LODIndex++;
	}

	StaticMesh->SetStaticMaterials(StaticMaterials);

	StaticMesh->InitResources();

	StaticMesh->CalculateExtendedBounds();

	return StaticMesh;
}

UUnrealSTLFactory::UUnrealSTLFactory()
{
	SupportedClass = UStaticMesh::StaticClass();
	Formats.Add(TEXT("stl;STL file"));
	bEditorImport = true;
}

UObject* UUnrealSTLFactory::FactoryCreateBinary(UClass* InClass, UObject* InParent, FName InName, EObjectFlags Flags, UObject* Context, const TCHAR* Type, const uint8*& Buffer, const uint8* BufferEnd, FFeedbackContext* Warn, bool& bOutOperationCanceled)
{
	UStaticMesh* StaticMesh = NewObject<UStaticMesh>(InParent, InName, Flags);

	FArrayReader Reader;
	Reader.Append(Buffer, BufferEnd - Buffer);

	FUnrealSTLMesh STLMesh;
	if (!LoadMeshFromSTLData(Reader, ImportConfig, STLMesh))
	{
		return nullptr;
	}

	UStaticMeshDescription* MeshDescription = UStaticMesh::CreateStaticMeshDescription();

	TVertexAttributesRef<FVector3f> Positions = MeshDescription->GetVertexPositions();
	TVertexInstanceAttributesRef<FVector3f> Normals = MeshDescription->GetVertexInstanceNormals();

	FPolygonGroupID PolygonGroup = MeshDescription->CreatePolygonGroup();

	for (int32 VertexIndex = 0; VertexIndex < STLMesh.LODIndices.Num(); VertexIndex += 3)
	{
		const uint32 VertexIndex1 = STLMesh.LODIndices[VertexIndex];
		const uint32 VertexIndex2 = STLMesh.LODIndices[VertexIndex + 1];
		const uint32 VertexIndex3 = STLMesh.LODIndices[VertexIndex + 2];

		const FVector3f Vertex1 = STLMesh.Vertices[VertexIndex1].Position;
		const FVector3f Normal1 = STLMesh.Vertices[VertexIndex1].TangentZ;

		const FVector3f Vertex2 = STLMesh.Vertices[VertexIndex2].Position;
		const FVector3f Normal2 = STLMesh.Vertices[VertexIndex2].TangentZ;

		const FVector3f Vertex3 = STLMesh.Vertices[VertexIndex3].Position;
		const FVector3f Normal3 = STLMesh.Vertices[VertexIndex3].TangentZ;

		const FVertexID VertexID1 = MeshDescription->CreateVertex();
		Positions[VertexID1] = Vertex1;
		FVertexInstanceID VertexInstanceID1 = MeshDescription->CreateVertexInstance(FVertexID(VertexIndex));
		Normals[VertexInstanceID1] = Normal1;

		const FVertexID VertexID2 = MeshDescription->CreateVertex();
		Positions[VertexID2] = Vertex2;
		FVertexInstanceID VertexInstanceID2 = MeshDescription->CreateVertexInstance(FVertexID(VertexIndex + 1));
		Normals[VertexInstanceID2] = Normal2;

		const FVertexID VertexID3 = MeshDescription->CreateVertex();
		Positions[VertexID3] = Vertex3;
		FVertexInstanceID VertexInstanceID3 = MeshDescription->CreateVertexInstance(FVertexID(VertexIndex + 2));
		Normals[VertexInstanceID3] = Normal3;

		TArray<FEdgeID> Edges;
		// fix winding when using mesh description TODO: add an import panel
		TArray<FVertexInstanceID> Instances = { VertexInstanceID1, VertexInstanceID3, VertexInstanceID2 };
		MeshDescription->CreatePolygon(PolygonGroup, Instances, Edges);
	}

	StaticMesh->BuildFromStaticMeshDescriptions({ MeshDescription }, false);

	return StaticMesh;
}
